/*-
 * Copyright (c) 2001 Jonathan Belson
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $FreeBSD$
 */


#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/syslog.h>
#include <sys/consio.h>
#include <sys/fbio.h>
#include <sys/random.h>

#include <dev/fb/fbreg.h>
#include <dev/fb/splashreg.h>
#include <dev/syscons/syscons.h>

#define MAX_X_SIZE	800
#define MAX_Y_SIZE	600
#define NUM_KIX		8
#define SPEED		12

#include "palette.c"

static int      blanked;

static u_char   buf[MAX_X_SIZE * (MAX_Y_SIZE + 1)];
static u_char	*vid;
static int		scrnmode, width, height, bpl;

typedef struct {
	int	x, y;
	int	x_inc, y_inc;
} point;

struct line { 
	point	p1, p2;
} lines[NUM_KIX];

static int head, tail;
static int hue;


/* Initialise each line */
static void
init_kix(void)
{
	int i;

	hue = 0;

	tail = 0;
	head = NUM_KIX - 1;

	for (i=0; i<NUM_KIX; i++) {
		lines[i].p1.x = 0;
		lines[i].p1.y = 0;
		lines[i].p2.x = 0;
		lines[i].p2.y = 0;

		lines[i].p1.x_inc = SPEED/2 + random()%(SPEED/2);
		lines[i].p1.y_inc = SPEED/2 + random()%(SPEED/2);
		lines[i].p2.x_inc = -SPEED/2 - random()%(SPEED/2);
		lines[i].p2.y_inc = -SPEED/2 - random()%(SPEED/2);
	}

	lines[head].p1.x = random()%width;
	lines[head].p1.y = random()%height;
	lines[head].p2.x = random()%width;
	lines[head].p2.y = random()%height;

}


/* Move specified line, bouncing off walls */
static void
step(int num)
{
	lines[num].p1.x += lines[num].p1.x_inc;
	lines[num].p1.y += lines[num].p1.y_inc;
	lines[num].p2.x += lines[num].p2.x_inc;
	lines[num].p2.y += lines[num].p2.y_inc;

	if (lines[num].p1.x < 0) {
		lines[num].p1.x = 0;
		lines[num].p1.x_inc *= -1;
	}
	if (lines[num].p1.x >= width) {
		lines[num].p1.x = width-1;
		lines[num].p1.x_inc *= -1;
	}
	if (lines[num].p2.x < 0) {
		lines[num].p2.x = 0;
		lines[num].p2.x_inc *= -1;
	}
	if (lines[num].p2.x >= width) {
		lines[num].p2.x = width-1;
		lines[num].p2.x_inc *= -1;
	}

	if (lines[num].p1.y < 0) {
		lines[num].p1.y = 0;
		lines[num].p1.y_inc *= -1;
	}
	if (lines[num].p1.y >= height) {
		lines[num].p1.y = height-1;
		lines[num].p1.y_inc *= -1;
	}
	if (lines[num].p2.y < 0) {
		lines[num].p2.y = 0;
		lines[num].p2.y_inc *= -1;
	}
	if (lines[num].p2.y >= height) {
		lines[num].p2.y = height-1;
		lines[num].p2.y_inc *= -1;
	}
}


/* Draw line in specified colour */
static void
draw_line(int x1, int y1, int x2, int y2, int colour)
{
	int		 dx, dy;
	u_char	*p;

	dx = x2 - x1;
	dy = y2 - y1;

	if (dx < 0)
		dx = -dx;
	if (dy < 0)
		dy = -dy;

	if (dx >= dy) {
		int d, x, y, x_end, y_inc;
		int inc1, inc2;

		if (x1 > x2) {
			x = x2;
			y = y2;
			x_end = x1;

			if (dy == 0) {
				y_inc = 0;
			} else if (y2 > y1) {
				y_inc = -bpl;
			} else {
				y_inc = bpl;
			}
		} else {
			x = x1;
			y = y1;
			x_end = x2;

			if (dy == 0) {
				y_inc = 0;
			} else if (y2 > y1) {
				y_inc = bpl;
			} else {
				y_inc = -bpl;
			}
		}

		inc1 = 2*dy;
		d = inc1 - dx;
		inc2 = 2*(dy - dx);

		p = buf + (bpl*y + x);
		*p = colour;

		while (x < x_end) {
			x++;
			p++;
			if (d < 0) {
				d += inc1;
			} else {
				p += y_inc;
				d += inc2;
			}
			*p = colour;
			
		}

	} else {
		int d, x, y, y_end, x_inc;
		int inc1, inc2;

		if (y1 > y2) {
			x = x2;
			y = y2;
			y_end = y1;

			if (dx == 0) {
				x_inc = 0;
			} else if (x2 > x1) {
				x_inc = -1;
			} else {
				x_inc = 1;
			}
		} else {
			x = x1;
			y = y1;
			y_end = y2;

			if (dx == 0) {
				x_inc = 0;
			} else if (x2 > x1) {
				x_inc = 1;
			} else {
				x_inc = -1;
			}
		}

		inc1 = 2*dx;
		d = inc1 - dy;
		inc2 = 2*(dx - dy);

		p = buf + (bpl*y + x);
		*p = colour;

		while (y < y_end) {
			y++;
			p += bpl;
			if (d < 0) {
				d += inc1;
			} else {
				p += x_inc;
				d += inc2;
			}
			*p = colour;
			
		}

	}
}

static int
kix_saver(video_adapter_t *adp, int blank)
{
	int i;

    if (blank) {
		if (blanked <= 0) {
			set_video_mode(adp, scrnmode);

			bpl = adp->va_line_width;
			blanked++;
			vid = (u_char *) adp->va_window;

			load_palette(adp, palette);

			memset(vid, 0, width*height);
			memset(buf, 0, width*height);

			init_kix();
		}

		/* Erase tail */
		draw_line(lines[tail].p1.x, lines[tail].p1.y, lines[tail].p2.x, lines[tail].p2.y, 0);

		/* Make the tail one step forward from the head */
		lines[tail] = lines[head];
		step(tail);
	
		tail = (tail + 1)%NUM_KIX;
		head = (head + 1)%NUM_KIX;

		hue = (hue + 1)%255;

		/* Draw lines */
		for (i=0;i<NUM_KIX; i++) {
			draw_line(lines[i].p1.x, lines[i].p1.y, lines[i].p2.x, lines[i].p2.y, 1+hue);
		}
		memcpy(vid, buf, width*height);

    } else {
		blanked = 0;
    }

    return 0;
}


static int
kix_initialise(video_adapter_t *adp)
{
    video_info_t    info;

	/* Find suitable screenmode */
	if (!get_mode_info(adp, M_VESA_CG640x480, &info)) {
		scrnmode = M_VESA_CG640x480;
	} else if (!get_mode_info(adp, M_VGA_CG320, &info)) {
		scrnmode = M_VGA_CG320;
	} else {
		log(LOG_NOTICE, "kix_saver: could not find suitable screen mode\n");
		return (ENODEV);
 	}
	width = info.vi_width;
	height = info.vi_height;
    blanked = 0;

    return 0;
}


static int
kix_terminate(video_adapter_t *adp)
{
    return 0;
}


static scrn_saver_t kix_module = {
    "kix_saver", kix_initialise, kix_terminate, kix_saver, NULL
};

SAVER_MODULE(kix_saver, kix_module);

